#include "config.h"
#include "gfxinterface.h"
#include "global.h"
#include "input.h"
#include "managers.h"
#include "menu.h"
#include "score.h"
#include "stars.h"
#include "time_util.h"
#include "xpms.h"
#include <cstdlib>
#include <iomanip>
#include <iostream>
#include <memory>
#include <stdexcept>
#include <string>
#include <vector>


using std::auto_ptr;


namespace { // No external linking


namespace Data {

// Arrival trajectories
const RCoord hang[] = {
	RCoord(512,-20), RCoord(512,512), RCoord(450,550), RCoord(400,550),
	RCoord(350,512), RCoord(350,450)
};
const RCoord b_dhang[] = {
	RCoord(256,1024+20), RCoord(256,400), RCoord(768,600), RCoord(768,400),
	RCoord(256,600)
};
const RCoord l_u[] = {
	RCoord(-20,312), RCoord(350,550), RCoord(700,550), RCoord(800,500)
};
const RCoord r_u[] = {
	RCoord(1024+20,312), RCoord(1024-350,550), RCoord(1024-700,550),
	RCoord(1024-800,500)
};
const RCoord u[] = {
	RCoord(256,-20), RCoord(256,600), RCoord(768,600), RCoord(768,200)
};
const RCoord l_loop[] = {
	RCoord(-20,512), RCoord(512,512), RCoord(552,552), RCoord(512,612),
	RCoord(462,552), RCoord(512,512), RCoord(552,552), RCoord(512,612),
	RCoord(462,552), RCoord(512,512)
};
const RCoord r_loop[] = {
	RCoord(1024+20,512), RCoord(512,512), RCoord(462,552), RCoord(512,612),
	RCoord(552,552), RCoord(512,512), RCoord(462,552), RCoord(512,612),
	RCoord(552,552), RCoord(512,512)
};
const RCoord left[] = {
	RCoord(-20,512), RCoord(512,512)
};
const RCoord right[] = {
	RCoord(1024+20,512), RCoord(512,512)
};
const RCoord sine[] = {
	RCoord(-20,512), RCoord(20,512), RCoord(40,512+8),RCoord(60,512+16),
	RCoord(80,512+32), RCoord(100,512+64), RCoord(120,512+64+32),
	RCoord(140,512+64+32+16), RCoord(160,512+64+32+16+8),
	RCoord(180,512+64+32+16+8), RCoord(200,512+64+32+16),
	RCoord(220,512+64+32), RCoord(240,512+64), RCoord(280,512+32),
	RCoord(300,512+16), RCoord(320,512+8), RCoord(340,512), RCoord(512,400)
};
const RCoord top_diag[] = {
	RCoord(-20,-20), RCoord(512,800)
};
const RCoord l_b_diag[] = {
	RCoord(-20,1024+20), RCoord(512,216)
};
const RCoord r_b_diag[] = {
	RCoord(1024+20,1024+20), RCoord(512,216)
};
const RCoord drop128[] = { RCoord(128,-20), RCoord(128,700) };
const RCoord drop256[] = { RCoord(256,-20), RCoord(256,700) };
const RCoord drop384[] = { RCoord(384,-20), RCoord(384,700) };
const RCoord drop512[] = { RCoord(512,-20), RCoord(512,700) };
const RCoord drop640[] = { RCoord(640,-20), RCoord(640,700) };
const RCoord drop768[] = { RCoord(768,-20), RCoord(768,700) };
const RCoord drop896[] = { RCoord(896,-20), RCoord(896,700) };

#define SIZETAB(array) sizeof (array) / sizeof (array)[0]

const WaveData wd1[] = {
	{  0, alien1_xpm, 50, hang, SIZETAB(hang), true },
	{100, alien2_xpm, 50, hang, SIZETAB(hang), true },
	{100, alien3_xpm, 50, hang, SIZETAB(hang), true },
	{100, alien4_xpm, 50, hang, SIZETAB(hang), true },
	{100, alien5_xpm, 50, hang, SIZETAB(hang), true },
};
const WaveData wd2[] = {
	{200, alien6_xpm, 50, l_u, SIZETAB(l_u), false },
	{150, alien1_xpm, 50, r_u, SIZETAB(r_u), false },
	{150, alien2_xpm, 50, l_u, SIZETAB(l_u), false },
	{150, alien3_xpm, 50, r_u, SIZETAB(r_u), false },
	{150, alien4_xpm, 50, hang, SIZETAB(hang), true },
};
const WaveData wd3[] = {
	{200, alien7_xpm, 50, u, SIZETAB(u), true },
	{150, alien7_xpm, 50, u, SIZETAB(u), true },
	{150, alien4_xpm, 50, u, SIZETAB(u), true },
	{150, alien5_xpm, 50, u, SIZETAB(u), true },
	{150, alien6_xpm, 50, l_u, SIZETAB(l_u), false },
};
const WaveData wd4[] = {
	{200, alien8_xpm, 50, l_loop, SIZETAB(l_loop), false },
	{ 25, alien8_xpm, 50, r_loop, SIZETAB(r_loop), false },
	{ 50, alien2_xpm, 50, l_b_diag, SIZETAB(l_b_diag), false },
	{ 50, alien1_xpm, 50, r_b_diag, SIZETAB(r_b_diag), false },
	{ 50, alien5_xpm, 50, b_dhang, SIZETAB(b_dhang), false },
};
const WaveData wd5[] = {
	{200, alien9_xpm, 50, hang, SIZETAB(hang), true },
	{ 50, alien9_xpm, 50, b_dhang, SIZETAB(b_dhang), true },
	{ 50, alien2_xpm, 50, l_b_diag, SIZETAB(l_b_diag), false },
	{ 50, alien1_xpm, 50, r_b_diag, SIZETAB(r_b_diag), false },
	{ 50, alien7_xpm, 50, b_dhang, SIZETAB(b_dhang), false },
};
const WaveData wd6[] = {
	{200, alien10_xpm, 50, sine, SIZETAB(sine), false },
	{  0, alien11_xpm, 50, top_diag, SIZETAB(top_diag), true },
	{ 50, alien1_xpm, 25, drop128, SIZETAB(drop128), true },
	{  5, alien2_xpm, 25, drop256, SIZETAB(drop256), true },
	{  5, alien3_xpm, 25, drop384, SIZETAB(drop384), true },
	{  5, alien4_xpm, 25, drop512, SIZETAB(drop512), true },
	{  5, alien5_xpm, 25, drop640, SIZETAB(drop512), true },
	{  5, alien6_xpm, 25, drop768, SIZETAB(drop512), true },
};

struct LevelData {
	const WaveData * wavesdata;
	int nwaves;
};

LevelData levels[] = {
	{ wd1, SIZETAB(wd1) },
	{ wd2, SIZETAB(wd2) },
	{ wd3, SIZETAB(wd3) },
	{ wd4, SIZETAB(wd4) },
	{ wd5, SIZETAB(wd5) },
	{ wd6, SIZETAB(wd6) },
};

} // namespace Data



/*
 *  Collisions functions
 */

int Collisions(AliensManager * aliens, BulletsManager * bullets,
               BonusManager * bonuses, ExplosionsManager * explosions)
{
	int nb_aliens_destroyed (0);
	BulletsManager::BulletsCtr & b_list (bullets->GetList());
	BulletsManager::BulletsCtr::iterator it (b_list.begin());
	while (it != b_list.end()) {
		const std::pair<Explosion*, Bonus*> boom (aliens->Collision(*it));
		if (boom.first) {
			explosions->Add(boom.first);
			++nb_aliens_destroyed;
			bonuses->AddIf(boom.second);
			it = bullets->Destroy(it);
		} else ++it;
		
	}
	return nb_aliens_destroyed;
}


void Collisions(Player * player, BulletsManager * bombs,
                ExplosionsManager * explosions)
{
	BulletsManager::BulletsCtr & b_list (bombs->GetList());
	BulletsManager::BulletsCtr::iterator it (b_list.begin());
	while (it != b_list.end()) {
	     	Explosion *const explosion (player->Collision(*it));
		if (explosion) {
			explosions->Add(explosion);
			it = bombs->Destroy(it);
		} else ++it;
	}
}


void Collisions(Player * player, BonusManager * bonuses)
{
	BonusManager::BonusCtr & b_list (bonuses->GetList());
	BonusManager::BonusCtr::iterator it (b_list.begin());
	while (it != b_list.end()) {
		if (player->Collision(*it)) it = bonuses->Destroy(it);
		else ++it;
	}
}



/*
 *  Main functions
 */

AliensManager * NewLevel(const Score * score)
{
	const size_t data_number ((score->Level()-1) % (SIZETAB(Data::levels)));
	return new AliensManager (score->Level(),
	                          Data::levels[data_number].wavesdata,
	                          Data::levels[data_number].nwaves,
		                  SIZETAB(Data::levels));
}


void Play(std::vector<StarsField> * stars_fields, Score * score)
{
	X11::Inst().ClearArea(Coord(0,0), X11::Inst().WindowWidth(), 
	                      X11::Inst().WindowHeight());
	Input input;
	Player player (score->Height() + 1);
	score->ReInit();
	score->SetShield(player.Shield());
	auto_ptr<AliensManager> aliens (NewLevel(score));
	BulletsManager bullets, bombs;
	BonusManager bonuses;
	ExplosionsManager explosions;
	int fast_star_scrolling (0);
        for (;;) {
		const double start_frame_time (CurrentMicroSecond());

		input.Update();
		if (input.Quit()) break;
		if (!input.Pause()) {
			Details::Instance().Change(input.Details());
			X11::Inst().GetWindowAttributes();
			
			score->Add(Collisions(aliens.get(), &bullets, &bonuses, &explosions));
			Collisions(&player, &bombs, &explosions);
			Collisions(&player, &bonuses);

			score->SetShield(player.Shield());
			if (player.Shield() < 0) break;

			if (aliens->Finished()) {
				score->IncLevel();
				auto_ptr<AliensManager> n (NewLevel(score));
				aliens = n;
				fast_star_scrolling = 100;
			}

			ScrollStarFields(stars_fields);
			if (fast_star_scrolling > 0) {
				--fast_star_scrolling;
				if (fast_star_scrolling == 0) score->EraseLevel();
				ScrollStarFields(stars_fields);
				ScrollStarFields(stars_fields);
				if (fast_star_scrolling > 0) score->ShowLevel();
			}

			aliens->Move();
			bonuses.Move();
			player.Move(input.Move());
			bombs.Move();
			bombs.Add(aliens->Fire(3 * score->Level() - bombs.Nb()));
			bullets.Move();
			if (input.Fire()) bullets.Add(player.Fire());
			explosions.Move();
			score->Refresh();
		}
		X11::Inst().Sync(False);
		WaitNextFrame(start_frame_time);
        }
}


void MainLoop()
{
	X11::CreateInstance();
	X11::Inst().SetWindowTitle("XGalaga++");
	
	// Background is made of scrolling stars.
	std::vector<StarsField> stars_fields;
	for (unsigned speed (1); speed <= 3; ++speed) {
		stars_fields.push_back(StarsField(speed));
	}

	Score score;

	while (StartMenu::Instance().Display(&stars_fields, &score)) {
		Play(&stars_fields, &score);
	}
}


} // namespace for no linking


int main()
{
	int code (0);
	try {
		MainLoop();
	} catch (const std::exception & e) {
		std::cerr << e.what() << std::endl;
		code = 1;
	}
	PixKeeper::DestroyInstance();
	StartMenu::DestroyInstance();
	X11::DestroyInstance();
	Details::DestroyInstance();
	return code;
}
